// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Narrow_impl.h>
#include <OB/Hashtable.h>
#include <OB/HashtableI.h>
#include <OB/Properties.h>
#include <OB/Hashers.h>
#include <OB/ExtNaming_skel.h>
#include <Types.h>

#include <ExtNaming_impl.h>

#include <NamingContextSet.h>

#include <stdio.h>
#include <NamingDatabase.h>

#include <time.h>

OB_MAKE_NARROW_IMPL_1(CosNaming_ExtNamingContext_impl,
                      CosNaming_ExtNamingContext_skel)

int CosNaming_ExtNamingContext_impl::count_ = 0;

CosNaming_BindingListenerSeq CosNaming_ExtNamingContext_impl::listeners_;

// ----------------------------------------------------------------------
// Template instantiations
// ----------------------------------------------------------------------

#ifndef HAVE_NO_EXPLICIT_TEMPLATES
template class OBHashtable<CORBA_String_var, CORBA_Object_var>;
template class OBHashtable<CosNaming_NameComponent, NcOrObjBinding*>;
#else
#ifdef HAVE_PRAGMA_DEFINE
#pragma define(OBHashtable<CORBA_String_var, CORBA_Object_var>)
#pragma define(OBHashtable<CosNaming_NameComponent, NcOrObjBinding*>)
#endif
#endif

// ----------------------------------------------------------------------
// Hashtable methods
// ----------------------------------------------------------------------

CORBA_ULong
OBHash(const CosNaming_NameComponent& n)
{
    return OBHash(n.id);
}

bool
OBHashCompare(const CosNaming_NameComponent& k1,
	      const CosNaming_NameComponent& k2)
{
    return (strcmp(k1.id, k2.id) == 0) && (strcmp(k1.kind, k2.kind) == 0);
}

// ----------------------------------------------------------------------
// NcOrObj class, and NcOrObjBinding
// ----------------------------------------------------------------------

class NcOrObj
{
    CORBA_Object_var obj_;
    CosNaming_NamingContext_var context_;
    CosNaming_BindingType type_;

public:

    NcOrObj()
    {
    }

    NcOrObj(CORBA_Object_ptr obj)
	: obj_(CORBA_Object::_duplicate(obj)),
	  type_(CosNaming_nobject)
    {
    }
    
    NcOrObj(CosNaming_NamingContext_ptr context)
	: context_(CosNaming_NamingContext::_duplicate(context)),
	  type_(CosNaming_ncontext)
    {
    }

    CORBA_Object_ptr obj() const
    {
	assert(type_ == CosNaming_nobject);
	return obj_;
    }

    CosNaming_NamingContext_ptr nc() const
    {
	assert(type_ == CosNaming_ncontext);
	return context_;
    }

    CosNaming_BindingType type() const
    {
	return type_;
    }
    
    CORBA_Object_ptr asObj() const
    {
	if(type_ == CosNaming_nobject)
	    return obj_;
	else
	    return context_;
    }
};

struct NcOrObjBinding
{
    NcOrObj ncOrObj;
    CosNaming_NameComponent comp;
    long timestamp;
};

// ----------------------------------------------------------------------
// CosNaming_NamingContext private member implementation
// ----------------------------------------------------------------------

void
CosNaming_ExtNamingContext_impl::doBind(const CosNaming_Name& n,
					const NcOrObj& ncOrObj,
					bool rebind,
					int timestamp)
{
    CORBA_ULong i;

    //
    // Check if name is valid
    //
    if(n.length() == 0)
	throw InvalidName();
    if(strlen(n[0].id) == 0 && strlen(n[0].kind) == 0)
	throw InvalidName();

    NcOrObjBinding* binding = 0;
    bindings_ -> get(n[0], binding);

    if(n.length() == 1)
    {
	//
	// Name is already bound
	//
	if(!rebind && binding)
	    throw AlreadyBound();

        if(store_)
            store_ -> bind(this, n, ncOrObj.asObj(), ncOrObj.type(), timestamp,
			   rebind);

	//
        // Bind the name and the object
	//
	if(!binding)
	{
	    binding = new NcOrObjBinding;
	    bindings_ -> put(n[0], binding);
	}

	binding -> comp = n[0];
	binding -> ncOrObj = ncOrObj;
	binding -> timestamp = timestamp;
    }
    else
    {
	//
	// Missing node if binding was not found
	//
	if(binding == 0)
	    throw NotFound(missing_node, n);

	//
	// Binding must belong to a context
	//
	if(binding -> ncOrObj.type() != CosNaming_ncontext)
	    throw NotFound(not_context, n);

	//
	// Context must not be a nil context here
	//
	if(CORBA_is_nil(binding -> ncOrObj.nc()))
	    throw CannotProceed(this, n);

	//
	// Create the rest of the name
	//
	CosNaming_Name rest;
	rest.length(n.length() - 1);
	for(i = 1 ; i < n.length() ; i++)
	    rest[i - 1] = n[i];

	//
	// Let the next context handle the binding
	//
	if(ncOrObj.type() == CosNaming_ncontext)
	{
	    if(rebind)
		binding -> ncOrObj.nc() -> rebind_context(rest, ncOrObj.nc());
	    else
		binding -> ncOrObj.nc() -> bind_context(rest, ncOrObj.nc());
	}
	else
	{
	    if(rebind)
		binding -> ncOrObj.nc() -> rebind(rest, ncOrObj.obj());
	    else
		binding -> ncOrObj.nc() -> bind(rest, ncOrObj.obj());
	}
    }
}

void
CosNaming_ExtNamingContext_impl::bindingAdded(const CosNaming_Name& n)
{
    //
    // Check whether change occured in this context
    //
    if(n.length() != 1)
	return;

    NcOrObjBinding* binding = 0;
    bindings_ -> get(n[0], binding);
    bool isContext = (binding -> ncOrObj.type() == CosNaming_ncontext);
    
    for(CORBA_ULong i = 0 ; i < listeners_.length() ; i++)
    {
	CosNaming_BindingListener_var l = listeners_[i];
	
	try
	{
	    if(isContext)
		l -> context_added(this, n[0]);
	    else
		l -> object_added(this, n[0]);
	}
	catch(const CORBA_COMM_FAILURE&)
	{
	    remove_client(l);
	}
    }
}

void
CosNaming_ExtNamingContext_impl::bindingRemoved(const CosNaming_Name& n,
						bool isContext)
{
    //
    // Check whether change occured in this context
    //
    if(n.length() != 1)
	return;

    for(CORBA_ULong i = 0 ; i < listeners_.length() ; i++)
    {
	CosNaming_BindingListener_var l = listeners_[i];

	try
	{
	    if(isContext)
		l -> context_removed(this, n[0]);
	    else
		l -> object_removed(this, n[0]);
	}
	catch(const CORBA_COMM_FAILURE&)
	{
	    remove_client(l);
	}
    }
}

void
CosNaming_ExtNamingContext_impl::bindingChanged(const CosNaming_Name& n,
						bool isContext)
{
    //
    // Check whether change occured in this context
    //
    if(n.length() != 1)
	return;
    
    for(CORBA_ULong i = 0 ; i < listeners_.length() ; i++)
    {
	CosNaming_BindingListener_var l = listeners_[i];

	try
	{
	    if(isContext)
		l -> context_changed(this, n[0]);
	    else
		l -> object_changed(this, n[0]);
	}
	catch(CORBA_COMM_FAILURE&)
	{
	    remove_client(l);
	}
    }
}

// ----------------------------------------------------------------------
// CosNaming_NamingContext constructor and destructor
// ----------------------------------------------------------------------

CosNaming_ExtNamingContext_impl::CosNaming_ExtNamingContext_impl(
    CORBA_ORB_ptr orb,
    NamingDatabase* store,
    NamingContextSet& ncs,
    const char* key)
    : orb_(CORBA_ORB::_duplicate(orb)),
      key_(CORBA_string_dup(key)),
      bindings_(new CosNaming_ExtNamingContext_impl::BindingsHash(1023)),
      ncs_(ncs),
      store_(store),
      destroyed_(false)
{
    orb_ -> connect(this, key_);
    ncs_.add(this);

    if(store_)
    {
	store_ -> info();
        store_ -> create(this, time(0));
    }

    const char* prop = OBProperties::instance() ->
			getProperty("ooc.naming.display_root_ior");

    if(prop != 0 && strcmp(prop, "true") == 0 &&
       strcmp(key, "DefaultNamingContext") == 0)
    {
        CORBA_String_var s = orb_ -> object_to_string(this);
        cout << s << endl;
    }
}

CosNaming_ExtNamingContext_impl::~CosNaming_ExtNamingContext_impl()
{
}

// ----------------------------------------------------------------------
// CosNaming_NamingContext public member implementation
// ----------------------------------------------------------------------

//
// Bind without timestamp (standard mapping)
//
void
CosNaming_ExtNamingContext_impl::bind(const CosNaming_Name& n,
				      CORBA_Object_ptr obj)
{
    bind(n, obj, time(0));
}

//
// Bind with timestamp
//
void
CosNaming_ExtNamingContext_impl::bind(const CosNaming_Name& n,
				      CORBA_Object_ptr obj, int timestamp)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    NcOrObj ncOrObj(obj);

    doBind(n, ncOrObj, false, timestamp);

    bindingAdded(n);
}

void
CosNaming_ExtNamingContext_impl::rebind(const CosNaming_Name& n,
					CORBA_Object_ptr obj)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();
    
    NcOrObj ncOrObj(obj);

    //
    // AlreadyBound must not be thrown here, so convert it.
    //
    try
    {
	doBind(n, ncOrObj, true, time(0));
    }
    catch(AlreadyBound&)
    {
	throw new CannotProceed(this, n);
    }
 
    bindingChanged(n, false);
}

//
// Bind without timestamp (standard mapping)
//
void
CosNaming_ExtNamingContext_impl::bind_context(const CosNaming_Name& n,
					      CosNaming_NamingContext_ptr nc)
{
    bind_context(n, nc, time(0));
}

//
// Bind with timestamp
//
void
CosNaming_ExtNamingContext_impl::bind_context(const CosNaming_Name& n,
					      CosNaming_NamingContext_ptr nc,
					      int timestamp)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();
    
    NcOrObj ncOrObj(nc);
    doBind(n, ncOrObj, false, timestamp);

    bindingAdded(n);
}

void
CosNaming_ExtNamingContext_impl::rebind_context(const CosNaming_Name& n,
						CosNaming_NamingContext_ptr nc)
{
    OB_SYNCHRONIZED
	
    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();
    NcOrObj ncOrObj(nc);

    //
    // AlreadyBound must not be thrown here, so convert it.
    //
    try
    {
	doBind(n, ncOrObj, true, time(0));
    }
    catch(AlreadyBound&)
    {
	throw new CannotProceed(this, n);
    }

    bindingChanged(n, true);
}

CORBA_Object_ptr
CosNaming_ExtNamingContext_impl::resolve(const CosNaming_Name& n)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    CORBA_ULong i;

    //
    // Check if name is valid
    //
    if(n.length() == 0)
	throw InvalidName();
    if(strlen(n[0].id) == 0 && strlen(n[0].kind) == 0)
	throw InvalidName();

    NcOrObjBinding* binding = 0;
    bindings_ -> get(n[0], binding);

    //
    // Missing node if binding was not found
    //
    if(binding == 0)
	throw NotFound(missing_node, n);

    //
    // Object or context found, return it
    //
    if(n.length() == 1)
	return CORBA_Object::_duplicate(binding -> ncOrObj.asObj());

    //
    // Binding must belong to a context
    //
    if(binding -> ncOrObj.type() != CosNaming_ncontext)
	throw NotFound(not_context, n);
    
    //
    // Context must not be a nil context here
    //
    if(CORBA_is_nil(binding -> ncOrObj.nc()))
	throw CannotProceed(this, n);
    
    //
    // Create the rest of the name
    //
    CosNaming_Name rest;
    rest.length(n.length() - 1);
    for(i = 1 ; i < n.length() ; i++)
	rest[i - 1] = n[i];
    
    //
    // Let the next context handle the resolving of the name
    //
    return binding -> ncOrObj.nc() -> resolve(rest);
}

void
CosNaming_ExtNamingContext_impl::unbind(const CosNaming_Name& n)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    CORBA_ULong i;

    //
    // Check if name is valid
    //
    if(n.length() == 0)
	throw InvalidName();
    if(strlen(n[0].id) == 0 && strlen(n[0].kind) == 0)
	throw InvalidName();

    NcOrObjBinding* binding = 0;
    bindings_ -> get(n[0], binding);

    //
    // Missing node if binding was not found
    //
    if(binding == 0)
	throw NotFound(missing_node, n);

    if(n.length() == 1)
    {
	//
	// Do unbind
	//
	bindings_ -> remove(n[0]);
        if(store_)
            store_ -> unbind(this, n);

	bindingRemoved(n, binding -> ncOrObj.type() == CosNaming_ncontext);
	delete binding;
    }
    else
    {
	//
	// Binding must belong to a context
	//
	if(binding -> ncOrObj.type() != CosNaming_ncontext)
	    throw NotFound(not_context, n);

	//
	// Context must not be a nil context
	//
	if(binding -> ncOrObj.nc() == 0)
	    throw CannotProceed(this, n);

        //
	// Create the rest of the name
	//
	CosNaming_Name rest;
	rest.length(n.length() - 1);
	for(i = 1 ; i < n.length() ; i++)
	    rest[i - 1] = n[i];

	//
	// Let the next context handle the resolving of the name
	//
	binding -> ncOrObj.nc() -> unbind(rest);
    }
}

CosNaming_NamingContext_ptr
CosNaming_ExtNamingContext_impl::new_context()
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    //
    // Create new unique key. The key is composed of the current time
    // and a running count.
    //
    CORBA_String_var name = CORBA_string_dup("");
    name += time(0);
    name += "-";
    name += count_;
    ++count_;

    CosNaming_NamingContext_ptr nc =
	new CosNaming_ExtNamingContext_impl(orb_, store_, ncs_, name);
    return nc;
}

CosNaming_NamingContext_ptr
CosNaming_ExtNamingContext_impl::bind_new_context(const CosNaming_Name& n)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    CosNaming_NamingContext_ptr nc = new_context();
    bind_context(n, nc);
    return nc;
}

void
CosNaming_ExtNamingContext_impl::destroy()
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();
    
    CosNaming_ExtNamingContext_impl::BindingsHash::Enumerator keys =
	bindings_ -> keys();
    if(keys.length() != 0)
	throw NotEmpty();

    if(store_)
        store_ -> destroy(this);

    //
    // Remove ourself from the NamingContextSet. This will stop the
    // naming database from dumping this context.
    //
    ncs_.remove(this);
    destroyed_ = true;

    //
    // Disconnect the servant from the ORB.
    //
    orb_ -> disconnect(this);
}

void
CosNaming_ExtNamingContext_impl::list(CORBA_ULong howMany,
				      CosNaming_BindingList*& bl,
				      CosNaming_BindingIterator_ptr& bi)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    CORBA_ULong i;

    CosNaming_ExtNamingContext_impl::BindingsHash::Enumerator keys =
	bindings_ -> keys();
    CORBA_ULong len = keys.length();
    CORBA_ULong num = len < howMany ? len : howMany;
	
    bl = new CosNaming_BindingList(num);
    bl -> length(num);

    for(i = 0 ; i < num ; i++)
    {
	NcOrObjBinding* binding;
	bindings_ -> get(keys[i], binding);
	assert(binding);
	(*bl)[i].binding_name.length(1);
	(*bl)[i].binding_name[0].id = binding -> comp.id;
	(*bl)[i].binding_name[0].kind = binding -> comp.kind;
	(*bl)[i].binding_type = binding -> ncOrObj.type();
    }

    if(howMany < len)
    {
	CosNaming_BindingList rest(len - howMany);
	rest.length(len - howMany);

	for(i = 0 ; i < len - howMany ; i++)
	{
	    NcOrObjBinding* binding;
	    bindings_ -> get(keys[i + howMany], binding);
	    assert(binding);
	    
	    rest[i].binding_name.length(1);
	    rest[i].binding_name[0].id = binding -> comp.id;
	    rest[i].binding_name[0].kind = binding -> comp.kind;
	    rest[i].binding_type = binding -> ncOrObj.type();
	}

	bi = new CosNaming_BindingIterator_impl(orb_, rest);
    }
    else
    {
	bi = CosNaming_BindingIterator::_nil();
    }
}

// ----------------------------------------------------------------------
// Extended naming context operations
// ----------------------------------------------------------------------

void
CosNaming_ExtNamingContext_impl::rename(const CosNaming_Name& oldName,
					const CosNaming_Name& newName)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    if(oldName.length() != 1 || newName.length() != 1)
	throw InvalidName();

    NcOrObjBinding* b = 0;
    bindings_ -> get(oldName[0], b);

    if(b == 0)
	throw NotFound(missing_node, oldName);
    if(strcmp(b -> comp.id, newName[0].id) ||
       strcmp(b -> comp.kind, newName[0].kind))
    {
	//
	// Check if new new already exists
	//
	if(bindings_ -> contains(newName[0]))
	    throw AlreadyBound();
	
	if(store_)
	    store_ -> unbind(this, oldName);
	
	b -> comp.id = newName[0].id;
	b -> comp.kind = newName[0].kind;
	bindings_ -> remove(oldName[0]);
	bindings_ -> put(newName[0], b);
	
	if(store_)
	    store_ -> bind(this, newName, b -> ncOrObj.asObj(),
			   b -> ncOrObj.type(), time(0), false);
	
	bindingChanged(newName, b -> ncOrObj.type() == CosNaming_ncontext);
    }
}

void
CosNaming_ExtNamingContext_impl::add_client(
    CosNaming_BindingListener_ptr client)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    for(CORBA_ULong i = 0 ; i < listeners_.length() ; i++)
    {
	if(client -> _is_equivalent(listeners_[i]))
	    return;
    }

    listeners_.append(CosNaming_BindingListener::_duplicate(client));
}

void
CosNaming_ExtNamingContext_impl::remove_client(
    CosNaming_BindingListener_ptr client)
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    for(CORBA_ULong i = 0 ; i < listeners_.length() ; i++)
    {
	if(client -> _is_equivalent(listeners_[i]))
	{
	    listeners_.remove(i);
	    return;
	}
    }
}

CosNaming_ExtNamingContext::ExtendedBindingList*
CosNaming_ExtNamingContext_impl::list_all()
{
    OB_SYNCHRONIZED

    if(destroyed_)
        throw CORBA_OBJECT_NOT_EXIST();

    CosNaming_ExtNamingContext_impl::BindingsHash::Enumerator keys =
	bindings_ -> keys();
    CORBA_ULong num = keys.length();

    ExtendedBindingList_var eb =
	new CosNaming_ExtNamingContext::ExtendedBindingList(num);
    eb -> length(num);

    for(CORBA_ULong i = 0 ; i < num ; i++)
    {
	NcOrObjBinding* binding;
	bindings_ -> get(keys[i], binding);
	assert(binding);

       eb[i].binding_name.length(1);
       eb[i].binding_name[0].id = binding -> comp.id;
       eb[i].binding_name[0].kind = binding -> comp.kind;
       eb[i].binding_type = binding -> ncOrObj.type();
       eb[i].timestamp = binding -> timestamp;
    }

    return eb._retn();
}

// ----------------------------------------------------------------------
// CosNaming_BindingIterator_impl constructor and destructor
// ----------------------------------------------------------------------
CosNaming_BindingIterator_impl::
CosNaming_BindingIterator_impl(CORBA_ORB_ptr orb,
			       const CosNaming_BindingList& bl)
    : orb_(CORBA_ORB::_duplicate(orb)), idx_(0), bl_(bl)
{
}

// ----------------------------------------------------------------------
// CosNaming_BindingIterator public member implementation
// ----------------------------------------------------------------------

CORBA_Boolean
CosNaming_BindingIterator_impl::next_one(CosNaming_Binding*& b)
{
    OB_SYNCHRONIZED
    if(idx_ < bl_.length())
    {
	b = new CosNaming_Binding(bl_[idx_++]);
	return true;
    }
    else
    {
	b = new CosNaming_Binding;
	b -> binding_type = CosNaming_nobject;
	return false;
    }
}

CORBA_Boolean
CosNaming_BindingIterator_impl::next_n(CORBA_ULong howMany,
				       CosNaming_BindingList*& bl)
{
    OB_SYNCHRONIZED

    CORBA_ULong len = bl_.length();
    CORBA_ULong num = len - idx_ < howMany ? len - idx_ : howMany;
    
    bl = new CosNaming_BindingList(num);
    bl -> length(num);
    
    for(CORBA_ULong i = 0 ; i < num ; i++)
	(*bl)[i] = bl_[idx_++];

    return howMany <= num;
}

void
CosNaming_BindingIterator_impl::destroy()
{
    OB_SYNCHRONIZED
    orb_ -> disconnect(this);
}
