// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Properties.h>
#include <OB/NTService.h>
#include <OB/Narrow_impl.h>
#include <OB/ExtNaming_skel.h>

#include <Types.h>

#include <ExtNaming_impl.h>
#include <NamingContextSet.h>

#include <stdio.h>
#include <stdlib.h>

#include <NamingDatabase.h>

#include <Logger.h>

#include <NTNamingService.h>

#include <NamingMsg.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

// ----------------------------------------------------------------------
// NTNamingService constructor
// ----------------------------------------------------------------------

NTNamingService::NTNamingService()
    : OBNTService("NamingService", "ORBacus Naming Service", NAMING_MSG),
      stopped_(false)
{
}

// ----------------------------------------------------------------------
// NTNamingService public member implementation
// ----------------------------------------------------------------------

void
NTNamingService::start(int argc, char** argv)
{
    try
    {
	//
	// Create ORB
	//
	CORBA_ORB_var orb = CORBA_ORB_init(argc, argv);

	//
	// Resolve the service object if available as an initial reference.
	//
	const char* service = "NameService";
	CORBA_Object_var serviceObject;
	try
	{
	    serviceObject = orb -> resolve_initial_references(service);
	}
	catch(CORBA_ORB::InvalidName&)
	{
	}
	
	//
	// Set the IIOP port based on the port defined in the IIOP profile
	// of the service object (if it exists). Then create the BOA.
	//
	CORBA_BOA::set_iiop_port_from_object(serviceObject);
	boa_ = orb -> BOA_init(argc, argv);
	
	orb -> conc_model(CORBA_ORB::ConcModelBlocking);
	boa_ -> conc_model(CORBA_BOA::ConcModelThreadPerClient);
	
	//
	// Get options
	//
	const int defaultTimeout = 5; // Every 5 minutes.
	int timeout = defaultTimeout;
	
	CORBA_String_var database;
	
	OBMessageViewer* viewer = OBMessageViewer::instance();

	//
	// Check properties first.
	//
	OBProperties* properties = OBProperties::instance();
	OBProperties::KeySeq keys = properties -> getKeys("ooc.naming.");
	CORBA_ULong j;
	for(j = 0; j < keys.length(); ++j)
	{
	    const char* key = keys[j];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
	    
	    if(strcmp(key, "ooc.naming.database") == 0)
	    {
		database = CORBA_string_dup(value);
	    }
	    else if(strcmp(key, "ooc.naming.timeout") == 0)
	    {
		timeout = atoi(value);
	    }
	    else
	    {
		CORBA_String_var msg = CORBA_string_dup("`");
		msg += key;
		msg += "': unknown property";
		viewer -> warning(msg);
	    }
	}
	
	if(timeout <= 0)
	{
	    viewer -> warning(
		"timeout value must be > 0. Using default value");
	    timeout = defaultTimeout;
	}
	
	NamingContextSet* ncs = new NamingContextSet;
	NamingDatabase* store = 0;
	Logger* logger = 0;
	
	bool start = false;
	if(database != 0)
	{
	    //
	    // Create a new NamingDatabase
	    //
	    store = new NamingDatabase(boa_, orb, database, *ncs, start, true);
	    
	    //
	    // Start logger
	    //
	    logger = new Logger(store, timeout*60);
	    logger -> start();
	}
	else
	{
	    viewer -> warning("Running in non-persistent mode.");
	}
        
	//
	// Create the root naming context, if we've just created the
        // database.
	//
	CosNaming_NamingContext_var root;
	if(start)
	    root = new CosNaming_ExtNamingContext_impl(orb, store, *ncs,
						       "DefaultNamingContext");
	
        //
        // Inform the status manager that the service is running.
        //
	statusUpdate(SERVICE_RUNNING);
        
	//
	// Run implementation
	//
        if(!stopped_)
            boa_ -> impl_is_ready(CORBA_ImplementationDef::_nil());
        boa_ = CORBA_BOA::_nil();
	logger -> halt();
    }
    catch(CORBA_SystemException& ex)
    {
	CORBA_String_var err = OBExceptionToString(ex);
	OBMessageViewer::instance() -> error(err);
    }
}

void
NTNamingService::stop()
{
    stopped_ = true;
    if(!CORBA_is_nil(boa_))
        boa_ -> deactivate_impl(CORBA_ImplementationDef::_nil());
}
