// **********************************************************************
//
// Copyright (c) 1999
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1999
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#include <OB/CORBA.h>
#include <OB/Util.h>
#include <OB/Timer.h>
#include <OB/Narrow_impl.h>
#include <OB/ExtNaming_skel.h>
#include <OB/Properties.h>
#include <Types.h>

#include <ExtNaming_impl.h>

#include <NamingContextSet.h>

#include <stdio.h>
#include <stdlib.h>

#include <NamingDatabase.h>
#include <Logger.h>

#include <Config.h> // Make sure that the compiler complains if the
                    // configuration doesn't match

static void
usage(const char* progName)
{
    cerr << "Usage:\n";
    cerr << progName << " [options]\n"
	"\n"
	"Options:\n"
	"-h, --help             Show this message.\n"
	"-v, --version          Show Naming Service version.\n"
	"-i, --ior              Print IOR on standard output.\n"
        "-s, --start            Run for the first time.\n"
        "-d, --database FILE    Use FILE as the database.\n"
        "-t, --timeout MINS     Purge the database every MINS minutes.\n"
	"                       (The default is 5 minutes.)\n"
	;
}

int
main(int argc, char* argv[], char*[])
{
    try
    {
	//
	// Create ORB
	//
	CORBA_ORB_var orb = CORBA_ORB_init(argc, argv);

	//
	// Resolve the service object if available as an initial reference.
	//
	const char* service = "NameService";
	CORBA_Object_var serviceObject;
	try
	{
	    serviceObject = orb -> resolve_initial_references(service);
	}
	catch(CORBA_ORB::InvalidName&)
	{
	}
	
	//
	// Set the IIOP port based on the port defined in the IIOP profile
	// of the service object (if it exists). Then create the BOA.
	//
	CORBA_BOA::set_iiop_port_from_object(serviceObject);
	CORBA_BOA_var boa = orb -> BOA_init(argc, argv);
	
#ifdef HAVE_JTC
	orb -> conc_model(CORBA_ORB::ConcModelBlocking);
	//
	// Ensure the BOA thread model is suitable.
	//
	CORBA_BOA::ConcModel model = boa -> conc_model();
        if(model != CORBA_BOA::ConcModelThreadPerClient &&
           model != CORBA_BOA::ConcModelThreadPool &&
           model != CORBA_BOA::ConcModelThreadPerRequest)
        {
	    boa -> conc_model(CORBA_BOA::ConcModelThreadPerClient);
        }
#else
	//
	// If we haven't compiled with JThreads/C++ we run in reactive
	// model.
	//
	orb -> conc_model(CORBA_ORB::ConcModelReactive);
	boa -> conc_model(CORBA_BOA::ConcModelReactive);
#endif
	
	//
	// Get options
	//
	bool ior = false;
	bool start = false;
	int timeout = 5; // Every 5 minutes.
	
	CORBA_String_var database;
	
	//
	// Check properties first.
	//
	OBProperties* properties = OBProperties::instance();
	OBProperties::KeySeq keys = properties -> getKeys("ooc.naming.");
	CORBA_ULong j;
	for(j = 0; j < keys.length(); ++j)
	{
	    const char* key = keys[j];
	    const char* value = properties -> getProperty(key);
	    assert(value != 0);
	    
	    if(strcmp(key, "ooc.naming.database") == 0)
	    {
		database = CORBA_string_dup(value);
	    }
	    else if(strcmp(key, "ooc.naming.timeout") == 0)
	    {
		timeout = atoi(value);
	    }
	    else
	    {
		cerr << argv[0] << ": `" << key << "': unknown property"
		     << endl;
	    }
	}
	
	CORBA_Long i;
	for(i = 1 ; i < argc && *argv[i] == '-' ; i++)
	{
	    if(strcmp(argv[i], "--help") == 0 ||
	       strcmp(argv[i], "-h") == 0)
	    {
		usage(argv[0]);
		return 0;
	    }
	    else if(strcmp(argv[i], "--version") == 0 ||
		    strcmp(argv[i], "-v") == 0)
	    {
		cerr << "Naming Service " << OBVersion << endl;
		return 0;
	    }
	    else if(strcmp(argv[i], "--ior") == 0 ||
		    strcmp(argv[i], "-i") == 0)
	    {
		ior = true;
	    }
	    else if(strcmp(argv[i], "--timeout") == 0 ||
		    strcmp(argv[i], "-t") == 0)
	    {
		++i;
		if(i >= argc)
		{
		    cerr << argv[0] << ": " << argv[i-1]
			 << " expects argument" << endl;
		    usage(argv[0]);
		    return 1;
		}
		timeout = atoi(argv[i]);
	    }
	    else if(strcmp(argv[i], "--database") == 0 ||
		    strcmp(argv[i], "-d") == 0)
	    {
		++i;
		if(i >= argc)
		{
		    cerr << argv[0] << ": " << argv[i-1]
			 << " expects argument" << endl;
		    usage(argv[0]);
		    return 1;
		}
		database = CORBA_string_dup(argv[i]);
	    }
	    else if(strcmp(argv[i], "--start") == 0 ||
		    strcmp(argv[i], "-s") == 0)
	    {
		start = true;
	    }
	    else
	    {
		cerr << argv[0] << ": unknown option `"
		     << argv[i] << "'\n" << endl;
		usage(argv[0]);
		return 1;
	    }
	}
	
	if(timeout <= 0)
	{
	    cerr << argv[0] << ": timeout value must be > 0" << endl;
	    return 1;
	}
	
	if(start && database == 0)
	{
	    cerr << argv[0]
		 << ": --start (-s) requires the the --database (-d) option"
		 << endl;
	    return 1;
	}
	
	//
	// Print IOR on standard output
	//
	properties -> setProperty("ooc.naming.display_root_ior",
				  (ior) ? "true" : "false");
	
	NamingContextSet* ncs = new NamingContextSet;
	NamingDatabase* store = 0;
	Logger* logger = 0;
	
	CosNaming_NamingContext_var root;
        
	if(database != 0)
	{
	    //
	    // Create a new NamingDatabase
	    //
	    store = new NamingDatabase(boa, orb, database, *ncs, start);
	    
	    //
	    // Start logger
	    //
	    logger = new Logger(store, timeout*60);
#ifdef HAVE_JTC
	    logger -> start();
#endif
	}
	else
	{
	    cerr << "Running in non-persistent mode." << endl;
	}
        
	//
	// Create the root naming context
	//
	if(start || database == 0)
	    root = new CosNaming_ExtNamingContext_impl(orb, store, *ncs,
						       "DefaultNamingContext");
	
	//
	// Run implementation
	//
	boa -> impl_is_ready(CORBA_ImplementationDef::_nil());

#ifdef HAVE_JTC
	if(logger)
	    logger -> halt();
#endif
    }
    catch(CORBA_SystemException& ex)
    {
	OBPrintException(ex);
	return 1;
    }

    return 0;
}
